/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: demux_mpeg_block_dxr2_dxr2.c,v 1.28 2001/08/05 00:59:50 ehasenle Exp $
 *
 * demultiplexer for mpeg 1/2 program streams with help from dxr2 dvd/vcd decoder
 * card. taken from deux_mpeg_block.c and expanded by raphael zulliger
 *
 */
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <errno.h>
#include <stdlib.h>
#include <linux/types.h>
#include "/usr/src/linux/include/linux/dxr2ioctl.h"
#include <curses.h>
#include <linux/limits.h>
#include <math.h>

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>

#include "xine_internal.h"
#include "monitor.h"
#include "demux.h"
#include "utils.h"

#include "dxr2init.c"

#define NUM_PREVIEW_BUFFERS 400

static uint32_t xine_debug;
int dxr2FD;

typedef struct demux_mpeg_block_dxr2_s {
  demux_plugin_t        demux_plugin;
  fifo_buffer_t        *audio_fifo;
  fifo_buffer_t        *video_fifo;
  input_plugin_t       *input;
  pthread_t             thread;
  int                   status;
  int                   blocksize;
  int                   send_end_buffers;
  gui_get_next_mrl_cb_t next_mrl_cb;
  gui_branched_cb_t     branched_cb;
  uint8_t              *scratch;
} demux_mpeg_block_dxr2_t ;



// ***********************************************************************************************
// while playing the movie this function is always called (i think no other is called during play)
// ***********************************************************************************************
static void demux_mpeg_block_dxr2_parse_pack (demux_mpeg_block_dxr2_t *this, int preview_mode) {
//  off_t *buf = NULL;

 buf_element_t *buf = NULL;
 // fill the buffer with contents
 buf = this->input->read_block (this->input, this->video_fifo, this->blocksize);

 // if dxr2 device is opened
 if ( dxr2FD != -1 ) {
	 // write the content of the buffer (a real vob stream) directly into the dxr2 card
	 write(dxr2FD, buf->content, buf->size);
 }
 // free the old buffer. important
 buf->free_buffer ( buf );
}

static void *demux_mpeg_block_dxr2_loop (void *this_gen) {

  buf_element_t *buf = NULL;
  demux_mpeg_block_dxr2_t *this = (demux_mpeg_block_dxr2_t *) this_gen;

  // debug
  printf ("__demux_mpeg_block_dxr2_loop\n");

  this->send_end_buffers = 1;

  do {

    demux_mpeg_block_dxr2_parse_pack(this, 0);

  } while (this->status == DEMUX_OK) ;

  /*
  printf ("demux_mpeg_block_dxr2: demux loop finished (status: %d)\n",
	  this->status);
  */

  this->status = DEMUX_FINISHED;

  if (this->send_end_buffers) {
    buf = this->video_fifo->buffer_pool_alloc (this->video_fifo);
    buf->type            = BUF_CONTROL_END;
    buf->decoder_info[0] = 0; /* stream finished */
    this->video_fifo->put (this->video_fifo, buf);
    
    if(this->audio_fifo) {
      buf = this->audio_fifo->buffer_pool_alloc (this->audio_fifo);
      buf->type            = BUF_CONTROL_END;
      buf->decoder_info[0] = 0; /* stream finished */
      this->audio_fifo->put (this->audio_fifo, buf);
    }

  }

  pthread_exit(NULL);

  return NULL;
}


static void demux_mpeg_block_dxr2_stop (demux_plugin_t *this_gen) {
  
  demux_mpeg_block_dxr2_t *this = (demux_mpeg_block_dxr2_t *) this_gen;
  void *p;
  buf_element_t *buf;

  // debug
  printf ("__demux_mpeg_block_dxr2_stop\n");

  if (this->status != DEMUX_OK) {
    printf ("__demux_mpeg_block_dxr2: stop...ignored\n");
    return;
  }

  this->send_end_buffers = 0;
  this->status = DEMUX_FINISHED;

  pthread_join (this->thread, &p);

  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  // because of compilation errors i've commented this. don't know if it is neccessary...
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//this->video_fifo->clear(this->video_fifo);
//if (this->audio_fifo)
//  this->audio_fifo->clear(this->audio_fifo);

  buf = this->video_fifo->buffer_pool_alloc (this->video_fifo);
  buf->type            = BUF_CONTROL_END;
  buf->decoder_info[0] = 1; /* forced */

  this->video_fifo->put (this->video_fifo, buf);

  if(this->audio_fifo) {
    buf = this->audio_fifo->buffer_pool_alloc (this->audio_fifo);
    buf->type            = BUF_CONTROL_END;
    buf->decoder_info[0] = 1; /* forced */
    this->audio_fifo->put (this->audio_fifo, buf);
  }
  
}

static int demux_mpeg_block_dxr2_get_status (demux_plugin_t *this_gen) {
  demux_mpeg_block_dxr2_t *this = (demux_mpeg_block_dxr2_t *) this_gen;

  // debug
  printf ("__demux_mpeg_block_dxr2_get_status\n");

  return this->status;
}

static void demux_mpeg_block_dxr2_start (demux_plugin_t *this_gen,
				    fifo_buffer_t *video_fifo, 
				    fifo_buffer_t *audio_fifo,
				    off_t pos,
				    gui_get_next_mrl_cb_t next_mrl_cb,
				    gui_branched_cb_t branched_cb) 
{

  demux_mpeg_block_dxr2_t *this = (demux_mpeg_block_dxr2_t *) this_gen;
  buf_element_t *buf;

  // debug
  printf ("__demux_mpeg_block_dxr2_start\n");

  this->video_fifo  = video_fifo;
  this->audio_fifo  = audio_fifo;
  this->next_mrl_cb = next_mrl_cb;
  this->branched_cb = branched_cb;

  pos /= (off_t) this->blocksize;
  pos *= (off_t) this->blocksize;

  /* 
   * send start buffer
   */

  buf = this->video_fifo->buffer_pool_alloc (this->video_fifo);
  buf->type    = BUF_CONTROL_START;
  this->video_fifo->put (this->video_fifo, buf);

  if(this->audio_fifo) {
    buf = this->audio_fifo->buffer_pool_alloc (this->audio_fifo);
    buf->type    = BUF_CONTROL_START;
    this->audio_fifo->put (this->audio_fifo, buf);
  }

  if((this->input->get_capabilities(this->input) & INPUT_CAP_SEEKABLE) != 0) {

    int num_buffers = NUM_PREVIEW_BUFFERS;

    this->input->seek (this->input, 0, SEEK_SET);

    this->status = DEMUX_OK ;
    while ( (num_buffers>0) && (this->status == DEMUX_OK) ) {

      demux_mpeg_block_dxr2_parse_pack(this, 1);
      num_buffers --;
    }

    printf (VERBOSE|DEMUX, "=>seek to %Ld\n",pos);
    this->input->seek (this->input, pos, SEEK_SET);
  }

  /*
   * query CLUT from the input plugin
   */

  buf = this->video_fifo->buffer_pool_alloc (this->video_fifo);

  if(this->input->get_optional_data(this->input, buf->mem,
     INPUT_OPTIONAL_DATA_CLUT) == INPUT_OPTIONAL_SUCCESS) {
    buf->type = BUF_SPU_CLUT;
    buf->content = buf->mem;

    this->video_fifo->put(this->video_fifo, buf);
  } else {
    buf->free_buffer(buf);
  }

  /*
   * now start demuxing
   */
  this->status = DEMUX_OK ;

  pthread_create (&this->thread, NULL, demux_mpeg_block_dxr2_loop, this) ;
}

static int demux_mpeg_block_dxr2_open(demux_plugin_t *this_gen,
				 input_plugin_t *input, int stage) {

  demux_mpeg_block_dxr2_t *this = (demux_mpeg_block_dxr2_t *) this_gen;

  switch(stage) {

  case STAGE_BY_CONTENT: {
    
  // debug
  printf ("__demux_mpeg_block_dxr2_open: STAGE_BY_CONTENT\n");
    if((input->get_capabilities(input) & INPUT_CAP_SEEKABLE) != 0) {
      
      this->blocksize = input->get_blocksize(input);

      if (!this->blocksize) {

	/* detect blocksize */
	input->seek(input, 2048, SEEK_SET);
	if (!input->read(input, this->scratch, 4)) 
	  return DEMUX_CANNOT_HANDLE;

	if (this->scratch[0] || this->scratch[1] 
	    || (this->scratch[2] != 0x01) || (this->scratch[3] != 0xba)) {

	  input->seek(input, 2324, SEEK_SET);
	  if (!input->read(input, this->scratch, 4)) 
	    return DEMUX_CANNOT_HANDLE;
	  if (this->scratch[0] || this->scratch[1] 
	      || (this->scratch[2] != 0x01) || (this->scratch[3] != 0xba)) 
	    return DEMUX_CANNOT_HANDLE;
	  this->blocksize = 2324;
	  
	} else
	  this->blocksize = 2048;
      }

      input->seek(input, 0, SEEK_SET);
      if (input->read(input, this->scratch, this->blocksize)) {

	if (this->scratch[0] || this->scratch[1] 
	    || (this->scratch[2] != 0x01) || (this->scratch[3] != 0xba))
	  return DEMUX_CANNOT_HANDLE;

	/* if it's a file then make sure it's mpeg-2 */
	if ( !input->get_blocksize(input) 
	     && ((this->scratch[4]>>4) != 4) )
	  return DEMUX_CANNOT_HANDLE;
	  
	this->input = input;
	return DEMUX_CAN_HANDLE;
      }	
    }
    return DEMUX_CANNOT_HANDLE;
  }
  break;

  case STAGE_BY_EXTENSION: {
    char *media;
    char *ending;
    char *MRL;
    
	// debug
    printf ("__demux_mpeg_block_dxr2_open:STAGE_BY_EXTENSION\n");

    MRL = input->get_mrl (input);
    
    media = strstr(MRL, "://");
    if(media) {
      if(!strncmp(MRL, "dvd", 3) || !strncmp(MRL, "d4d", 3)
	 || (((!strncmp(MRL, "stdin", 5) || !strncmp(MRL, "fifo", 4))
	      && (!strncmp((media+3), "mpeg2", 5) ))) 
	 ) {
	this->blocksize = 2048;
	this->input = input;
	return DEMUX_CAN_HANDLE;
      }
      if(!strncmp(MRL, "vcd", 3)) {
	this->blocksize = 2324;
	this->input = input;
	return DEMUX_CAN_HANDLE;
      }
    } 
    
    /*
     * check ending
     */
    
    ending = strrchr(MRL, '.');
    
    printf(VERBOSE|DEMUX, "demux_mpeg_block_dxr2_can_handle: ending %s of %s\n",
	    ending ? ending :"(none)", MRL);
    
    if(!ending)
      return DEMUX_CANNOT_HANDLE;
    
    if(!strcasecmp(ending, ".vob")) {
      this->blocksize = 2048;
      this->input = input;
      return DEMUX_CAN_HANDLE;
    }
  }
  break;

  default:
    return DEMUX_CANNOT_HANDLE;
    break;
  }

  return DEMUX_CANNOT_HANDLE;
}

static char *demux_mpeg_block_dxr2_get_id(void) {
  // debug
  printf ("__demux_mpeg_block_dxr2_get_id\n");

  return "MPEG_BLOCK_DXR2";
}

static void demux_mpeg_block_dxr2_close (demux_plugin_t *this) {
  // debug
  printf ("__demux_mpeg_block_dxr2_close\n");

  // close it again
  if(dxr2FD != -1) {
    printf("play status = %i\n", ioctl(dxr2FD, DXR2_IOC_STOP, NULL));
    close(dxr2FD);
  }
}

demux_plugin_t *init_demuxer_plugin(int iface, config_values_t *config) {

  demux_mpeg_block_dxr2_t *this;

  // debug
  printf ("__init_demuxer_plugin\n");

  if (iface != 2) {
    printf( "demux_mpeg: plugin doesn't support plugin API version %d.\n"
	    "demux_mpeg: this means there's a version mismatch between xine and this "
	    "demux_mpeg: demuxer plugin.\nInstalling current input plugins should help.\n",
	    iface);
    return NULL;
  }

  // **********************
  // initializing dxr2-card
  // **********************
  dxr2FD = xine_dxr2_init( );

  this = xmalloc (sizeof (demux_mpeg_block_dxr2_t));
  xine_debug  = config->lookup_int (config, "xine_debug", 0);

  this->demux_plugin.interface_version = DEMUXER_PLUGIN_IFACE_VERSION;
  this->demux_plugin.open              = demux_mpeg_block_dxr2_open;
  this->demux_plugin.start             = demux_mpeg_block_dxr2_start;
  this->demux_plugin.stop              = demux_mpeg_block_dxr2_stop;
  this->demux_plugin.close             = demux_mpeg_block_dxr2_close;
  this->demux_plugin.get_status        = demux_mpeg_block_dxr2_get_status;
  this->demux_plugin.get_identifier    = demux_mpeg_block_dxr2_get_id;
  
  this->scratch = xmalloc_aligned (512, 4096);
    
  return (demux_plugin_t *) this;
}
